﻿using System.Collections;
using System.Collections.Generic;
using UnityEngine;
using System;

public class TurnManager : Singleton<TurnManager> {
	
    public EventHandler<EventArgs<PreparationArgs>> onCastingDone;

    public Vector2 castDurationThresholds_Percentile = new Vector2(0.1f, 0.9f);
    private Vector2 castDurationThresholds { get { return castDurationThresholds_Percentile * recordingTime; } }
    private bool paused = false;
    public float phaseTime = 1;
    private float recordingTime { get { return phaseTime * 0.95f; } }
    public float frequencyCenter_Low = 300;
    public float frequencyCenter_Mid = 450;
    public float frequencyCenter_High = 600;
    public float speedMultiplierMax = 3;
    public void Toggle(bool on) { this.paused = on; }

    private Player currentPlayer = Player.Left;

    protected override void Initialize()
    {
        StartCoroutine("PhaseLoop");
    }

    IEnumerator PhaseLoop()
    {
        while (!DSP.initialized || !GameplayManager.inGame)
            yield return new WaitForEndOfFrame();

        float power = 0;
        Pitch pitch = Pitch.Mid;

        while (true)
        {
            // unless we're not working
            while (paused) yield return new WaitForEndOfFrame();

            // Ask for a recording (give some space between recording and phase
            DSP.Record(currentPlayer, phaseTime, recordingTime, (duration, frequency) => {
                power = GetPowerFromDuration(duration);
                pitch = GetPitchFromFrequency(frequency);
            });

            float overrideDuration = 0;
            float overrideFrequency = 0;
            for (float t = 0; t < phaseTime; t += Time.deltaTime)
            {
                if ((Input.GetKey(KeyCode.Q) && currentPlayer == Player.Left) ||
                    (Input.GetKey(KeyCode.O) && currentPlayer == Player.Right))
                {
                    overrideDuration += Time.deltaTime;
                    overrideFrequency += Time.deltaTime * frequencyCenter_High;
                }
                else if ((Input.GetKey(KeyCode.A) && currentPlayer == Player.Left) ||
                    (Input.GetKey(KeyCode.K) && currentPlayer == Player.Right))
                {
                    overrideDuration += Time.deltaTime;
                    overrideFrequency += Time.deltaTime * frequencyCenter_Mid;
                }
                else if ((Input.GetKey(KeyCode.Z) && currentPlayer == Player.Left) ||
                    (Input.GetKey(KeyCode.M) && currentPlayer == Player.Right))
                {
                    overrideDuration += Time.deltaTime;
                    overrideFrequency += Time.deltaTime * frequencyCenter_Low;
                }
                yield return new WaitForEndOfFrame();
            }
            Debug.Log(overrideFrequency + " : " + overrideDuration);
            if (overrideDuration > 0)
            {
                power = GetPowerFromDuration(overrideDuration);
                pitch = GetPitchFromFrequency(overrideFrequency / overrideDuration);
            }

            // Fire event
            Cast(power, pitch);

            // Swap players
            currentPlayer = currentPlayer == Player.Left ? Player.Right : Player.Left;
        }
    }

    private Pitch GetPitchFromFrequency(float frequency)
    {
        // Fq -> FMin
        if (frequency < (frequencyCenter_Low + frequencyCenter_Mid) / 2)
            return Pitch.Low;
        // Fq -> FMid
        else if (frequency < (frequencyCenter_Mid + frequencyCenter_High) / 2)
            return Pitch.Mid;
        else
            return Pitch.High;
    }

    private float GetPowerFromDuration(float duration)
    {
        if (duration < castDurationThresholds.x)
            return 0;
        else
        {
            return duration.Retargeted(castDurationThresholds, new Vector2(1f / speedMultiplierMax, 1));
        }
    }

    void Cast (float power, Pitch pitch)
    {
        string msg = "Casting for player " + currentPlayer + " done.\n";
        if (power > 0)
            msg += "Firing with " + power.PercentileToPercent() + " power at " + pitch + " pitch.";
        else
            msg += "No input this turn!";

        // this.Log(msg, LogType.Log, phaseTime);

        if (onCastingDone != null)
            onCastingDone(this, new EventArgs<PreparationArgs>(
                new PreparationArgs(currentPlayer, power, pitch)));
    }
}

public enum Player { Left = 0 , Right }
public enum Pitch { High, Mid, Low }

public struct PreparationArgs
{
    public PreparationArgs(Player player, float power, Pitch pitch)
    { this.player = player; this.power = power; this.pitch = pitch; }

    public Player player;
    public float power;
    public Pitch pitch;
}