﻿using System.Collections;
using System.Collections.Generic;
using UnityEngine;
using UnityEngine.SceneManagement;

public class MenuController : MonoBehaviour {

	public static MenuController instance;

	public Animator cameraAnimator;
	public Animator uiAnimator;

	public Animator menuAnimator;
    private AudioSource audioSource;
    private AudioClip musicMenu, musicGame;
	private Coroutine startCoroutine;
    private GameplayManager gameplayManager;
    public float musicVolume = 0.1f;
	void Start()
	{
		instance = this;
        musicMenu = The_Helper.ResourceLoad<AudioClip>("District Four");
        musicGame = The_Helper.ResourceLoad<AudioClip>("RetroFuture Nasty");
        audioSource = gameObject.AddComponentIfNotExists<AudioSource>();
        gameplayManager = gameObject.GetComponent<GameplayManager>();
        audioSource.clip = musicMenu;
        audioSource.Play();
        audioSource.volume = 0;
        audioSource.InterpolateVolume(musicVolume, .65f, false);

        List<string> names = new List<string>();
        foreach (string s in Microphone.devices)
            if (!s.ContainsInvariant("realtek") && !s.ContainsInvariant("built"))
                names.Add(s);

        if (names.Count == 0)
            names = new List<string>(Microphone.devices);

        mS = new MicrophoneSource(names.GetRandom(), 1, 44100);
        mS.Start();
    }
    MicrophoneSource mS;
    float currentVol = 0;
    void Update()
    {
        if (startCoroutine != null) return;

        if (Microphone.GetPosition(mS.deviceName) > 0)
        {
            // this.Log("Now recording.. " + (t / recordDuration).PercentileToPercent(), LogType.Log, Time.deltaTime * 2);
            float[] data = new float[1024];
            mS.source.GetOutputData(data, 0);
            float volume = 0;
            foreach (float d in data)
                volume += d * d;
            volume /= data.Length;
            volume = Mathf.Sqrt(volume);
            currentVol = Mathf.Lerp(currentVol, volume, .1f);
            // Debug.Log(currentVol);
            if (currentVol >= DSP.instance.voiceAcknowledgementThreshold * 3)
            {
                mS = null;
                startCoroutine = StartCoroutine(StartEnterGameAnimation());
            }
        }
        if (  Input.GetKeyUp(KeyCode.Space))
		{
			startCoroutine = StartCoroutine( StartEnterGameAnimation() );
		}

	}

	/*
	private IEnumerator TestRun()
	{
		while(Application.isPlaying)
		{
			yield return new WaitForSeconds( 5.0f );

			yield return StartEnterGameAnimation() ;

			yield return new WaitForSeconds( 5.0f );

			yield return CooroutineStartGameExitAnimation() ;
		}

	}
	*/

	public IEnumerator StartEnterGameAnimation()
	{
		menuAnimator.SetTrigger("Enter");
		yield return new WaitForSeconds(1f);
		StartCoroutine( StartCameraPanAnimation() );
	}

	public IEnumerator StartCameraPanAnimation()
    {
        cameraAnimator.SetTrigger("Enter");

        float cameraAnimDuration = 1.75f;
        audioSource.InterpolateVolume(0, cameraAnimDuration / 2, true);
        The_Helper.StartTimer(cameraAnimDuration / 2, a => {
            if (!audioSource) return;
            audioSource.clip = musicGame; 
            audioSource.Play();
        });

        float conveyerHeartsAnimDuration = 5/6f;
        The_Helper.StartTimer(cameraAnimDuration + conveyerHeartsAnimDuration, a =>
        {
            if (gameplayManager)
                gameplayManager.StartGameCountdown();
        });

        yield return new WaitForSeconds(cameraAnimDuration);
		uiAnimator.SetTrigger("Enter");
	}


	public void StartGameExitAnimation()
	{
		StartCoroutine( CooroutineStartGameExitAnimation() );
	}

	private IEnumerator CooroutineStartGameExitAnimation()
	{
		// SetToIdle
		//menuAnimator.SetTrigger("Idle");
		uiAnimator.SetTrigger("Exit");
		yield return new WaitForSeconds(1f);
		cameraAnimator.SetTrigger("Exit");
		yield return new WaitForSeconds(3f);

		menuAnimator.SetTrigger("Winner");
		yield return new WaitForSeconds(3f);

		SceneManager.LoadScene(0);
			
	}


}
