﻿using System;
using System.Collections;
using System.Collections.Generic;
using UnityEngine;
using UnityEngine.UI;
using UniRx;

using Shout.Input;

public class GameplayManager : MonoBehaviour, MicrophoneInput {

	public IObservable<PreparationArgs> DataStream{get; private set;}

	public GameObject player1Prefab;
	public GameObject player2Prefab;

	public Transform player1StartTransform;
	public Transform player2StartTransform;

	public PotionConveyorController player1Conveyor;
	public PotionConveyorController player2Conveyor;

    public HealthManager player1HealthManager, player2HealthManager;

	public Text MessageLabel;

    public Transform player1_OutputHigh, player1_OutputMid, player1_OutputLow;
    public Transform player2_OutputHigh, player2_OutputMid, player2_OutputLow;

    public Transform player1_InstrumentHigh, player1_InstrumentMid, player1_InstrumentLow;
    public Transform player2_InstrumentHigh, player2_InstrumentMid, player2_InstrumentLow;

    public AudioSource bardOffSource;

    public static bool inGame = false;

    void Start(){
        inGame = false;

        SetupTurnListener();
		InstantiatePlayers();
	}

	void InstantiatePlayers()
	{
		GameObject player1 =  GameObject.Instantiate( player1Prefab );
		PlayerController pController1 = player1.GetComponent<PlayerController>();
		pController1.player = Player.Left;
		pController1.transform.position = player1StartTransform.position;
        pController1.transform.rotation = player1StartTransform.rotation;
        pController1.transform.SetLossyScale(player1StartTransform.lossyScale);
        pController1.potionConveyorController = player1Conveyor;
		pController1.input = this;

		pController1.Initialize(player1HealthManager,
            player1_OutputHigh, player1_OutputMid, player1_OutputLow,
            player1_InstrumentHigh, player1_InstrumentMid, player1_InstrumentLow,
            The_Helper.ResourceLoad<Sprite>("Player1_Neck"),
            The_Helper.ResourceLoad<Sprite>("Player1_MouthOpen"),
            The_Helper.ResourceLoad<Sprite>("Player1_MouthClosed"),
            The_Helper.ResourceLoad<Sprite>("Player1_MouthErgh"));

		GameObject player2 =  GameObject.Instantiate( player2Prefab );
		PlayerController pController2 = player2.GetComponent<PlayerController>();
		pController2.player = Player.Right;
		pController2.transform.position = player2StartTransform.position;
		pController2.transform.rotation = player2StartTransform.rotation;
        pController2.transform.SetLossyScale(player2StartTransform.lossyScale);
        pController2.potionConveyorController = player2Conveyor;
		pController2.input = this;

		pController2.Initialize(player2HealthManager,
            player2_OutputHigh, player2_OutputMid, player2_OutputLow,
            player2_InstrumentHigh, player2_InstrumentMid, player2_InstrumentLow,
            The_Helper.ResourceLoad<Sprite>("Player2_Neck"),
            The_Helper.ResourceLoad<Sprite>("Player2_MouthOpen"),
            The_Helper.ResourceLoad<Sprite>("Player2_MouthClosed"),
            The_Helper.ResourceLoad<Sprite>("Player2_MouthErgh"));

		pController2.playerIsDeadCallback = delegate (Player player){
			Destroy( pController1);
			Destroy( pController2 );

			MenuController.instance.StartGameExitAnimation();
		};

		pController1.playerIsDeadCallback = delegate (Player player){
			Destroy( pController1);
			Destroy( pController2 );

			MenuController.instance.StartGameExitAnimation();
		};
	}

    internal void StartGameCountdown()
    {
        StartCoroutine(StartUpSequence());
    }

    private void Update()
    {
        if (Input.GetKey(KeyCode.LeftShift) && Input.GetKeyUp(KeyCode.R))
            UnityEngine.SceneManagement.SceneManager.LoadScene(0, UnityEngine.SceneManagement.LoadSceneMode.Single);
    }
    void SetupTurnListener()
	{
		DataStream = new Subject<PreparationArgs>();
	}

	void ShowMessage( string message )
	{
		MessageLabel.text = message;
	}
    public 
	IEnumerator StartUpSequence()
	{
        bardOffSource.Play();

        ShowMessage("3!");
		yield return new WaitForSeconds( 1f );
		ShowMessage("3! 2!");
		yield return new WaitForSeconds( 1f );
		ShowMessage("3! 2! 1!");
		yield return new WaitForSeconds( 1f );

		ShowMessage("Let the\nBard-Off\nBegin!!!");

		TurnManager.instance.onCastingDone +=OnPreperation;

		yield return new WaitForSeconds( 2f );
		ShowMessage("");
        inGame = true;
	}

	void OnPreperation(object sender, EventArgs<PreparationArgs> args){
        if (!inGame) return;
		// Debug.Log(string.Format("Player {0} input {1} power {2}", args.value.player, args.value.pitch.ToString(), args.value.power ));
		((Subject<PreparationArgs>)DataStream).OnNext(args.value);
	}

}
