﻿using System;
using System.Collections;
using System.Collections.Generic;
using UnityEngine;

public enum Potions{
	Nature = 0,
	Fire,
	Water
}

public class LaserLogic : MonoBehaviour {
	static Dictionary<Potions, Potions> defeats = null;

	public Player player;
	public float power = 0;
    public float speed = 10;
	private SpriteRenderer spriteRenderer;
    public Pitch pitch;
    public Transform effectsTransform;

	public Potions waveType;

	// Use this for initialization
	void Start () {
		if( defeats == null )
		{
			defeats = new Dictionary<Potions, Potions>();
			defeats.Add(Potions.Nature, Potions.Water);
			defeats.Add(Potions.Water, Potions.Fire);
			defeats.Add(Potions.Fire, Potions.Nature);
		}

		spriteRenderer = GetComponent<SpriteRenderer>();
		UpdateGraphics();
	}

    public static Color GetColorOfPotion(Potions potion)
    {
        switch (potion)
        {
            case Potions.Water:
                return Color.blue;

            case Potions.Fire:
                return Color.red;

            case Potions.Nature:
                return Color.green;
        }

        return Color.black;
    }

	void UpdateGraphics()
	{
        spriteRenderer.color = GetColorOfPotion(waveType);

    }
	
	// Update is called once per frame
	void Update () {
		if( power > 0 )
		{
			this.transform.position +=  this.transform.right * power * Time.deltaTime * speed;
            
		}
        effectsTransform.localPosition = SetEffectPosition(waveType, power);
    }

    private Vector3 SetEffectPosition(Potions waveType, float normalizedSpeed)
    {
        Vector3 localPos = Vector3.zero;
        float f = .35f;// normalizedSpeed; // GetEffectFrequency(pitch);
        float T = 1 / f;
        float A = 0.2f;// * normalizedSpeed;
        float loopedTime = Time.time % T;
        float yT = 0;
        // Triangle
        if (waveType == Potions.Fire)
        {
             yT = (loopedTime.Fold(T / 2) / (T / 2) - .25f) * A;
        }
        // Sin
        else if (waveType == Potions.Water)
        {
            yT = Mathf.Cos(2 * Mathf.PI * f * loopedTime) * A / 4;
        }
        // Square
        else if (waveType == Potions.Nature)
        {
            if (loopedTime < T / 2)
                yT = A / 4;
            else
                yT = -A / 4;
        }

        return Vector3.up * yT;
    }

    private float GetEffectFrequency(Pitch pitch)
    {
        if (pitch == Pitch.High)
            return 1.5f;
        else if (pitch == Pitch.Mid)
            return 1f;
        else if (pitch == Pitch.Low)
            return 0.5f;

        return 1;
    }

    void OnTriggerEnter2D( Collider2D col )
	{
        PlayerController pC = col.transform.parent ? 
            col.transform.parent.GetComponent<PlayerController>() : null;
        if (pC)
        {
            if (pC.player != player)
            {
                pC.Damage();
                Explode();
                return;
            }
        }
		var laserLogic = col.gameObject.GetComponent<LaserLogic>();
		if ( laserLogic != null )
		{
			if(	laserLogic.player != this.player )
			{
				LaserLogic.HandleInteract (laserLogic, this);
			}
		}

	}

    private void Explode(float scale = 1f)
    {
        string explosionName = waveType + "Explosion";
        if (scale < .5f)
            explosionName += "_0.1";
        GameObject explosion = gameObject.
            ResourceInstantiateAsChild_GameObject(explosionName);
        explosion.transform.parent = null;

        AudioSource.PlayClipAtPoint(The_Helper.ResourceLoad<AudioClip>(
            waveType + "Explosion"), transform.position, GetExplosionVolume(waveType) * scale);

        Destroy(explosion, 1);
        Destroy(gameObject);
    }

    private float GetExplosionVolume(Potions waveType)
    {
        if (waveType == Potions.Fire)
            return 0.5f;
        else if (waveType == Potions.Nature)
            return 1f;
        else if (waveType == Potions.Water)
            return 0.2f;

        return 1;
    }

    static void HandleInteract(  LaserLogic laser1, LaserLogic laser2 )
	{
		if( laser1.waveType == laser2.waveType )
		{
            laser1.Explode(0.1f);
            laser2.Explode(0.1f);
        }
        else{
			if( defeats[laser1.waveType] == laser2 .waveType)
            {
                laser2.Explode(0.1f);
            }
            else
            {
                laser1.Explode(0.1f);
			}
		}
	}

    void OnBecameInvisible()
	{
		Destroy(this.gameObject);
	}
}
