﻿using System.Collections;
using System.Collections.Generic;
using System.Linq;
using UnityEngine;

using UnityEngine.UI;

public class PotionConveyorController : MonoBehaviour {

	private Potions nextPotion;

	private Potions currentPotion;

	public GameObject NatureSprite;
	public GameObject FireSprite;
	public GameObject WaterSprite;

	private int[] potionTypeCount;

	public Animator getInAnimator;
	public Animator moveToAnimator;

	public Image SuckFillImage;
	public Image SuckCurrentImage;

	void Start()
	{
		potionTypeCount = new int[3];
		potionTypeCount[0] = 1;
		potionTypeCount[1] = 1;
		potionTypeCount[2] = 1;

		currentPotion = CreateRandomPotion();
		nextPotion = CreateRandomPotion();

		GameObject potionSprite = CreatePotionSprite( currentPotion );
		potionSprite.transform.SetParent( moveToAnimator.transform );
		potionSprite.transform.localScale = Vector3.one;
		potionSprite.transform.localPosition = Vector3.zero;

		potionSprite = CreatePotionSprite( nextPotion );
		potionSprite.transform.SetParent( getInAnimator.transform );
		potionSprite.transform.localScale = Vector3.one;
		potionSprite.transform.localPosition = Vector3.zero;

		//UpdateGraphics( currentPotionSprite, currentPotion );
		//UpdateGraphics( nextPotionSprite, nextPotion );
        lastPotions.Capacity = 10;
    }

	public Potions DrinkPotion()
	{
		var temp = currentPotion;
		currentPotion = nextPotion;

		nextPotion = CreateRandomPotion();

		//UpdateGraphics( currentPotionSprite, currentPotion );
		//UpdateGraphics( nextPotionSprite, nextPotion );

		StartCoroutine( StartConveyorAnimation(nextPotion) );
		StartCoroutine( SuckPotionAnimationCooroutine(temp) );

		return temp;
	}

	private void UpdateGraphics(Image renderer, Potions potion)
	{
		switch( potion ){		

		case Potions.Water:
			renderer.color = Color.blue;
		break;

		case Potions.Fire:
			renderer.color = Color.red;
			break;

		case Potions.Nature:
			renderer.color = Color.green;
			break;
		}

	}
    private List<Potions> lastPotions = new List<Potions>();
    private Potions CreateRandomPotion()
    {
        Potions potion = The_Helper.EnumGetRandom<Potions>();

        if (lastPotions.Count >= 2)
        {
            float[] occurencies = new float[The_Helper.EnumCount<Potions>()];
            foreach (Potions p in lastPotions)
                occurencies[(int)p]++;

            float[] probabilities = new float[The_Helper.EnumCount<Potions>()];

            for (int i = 0; i < probabilities.Length; i++)
                probabilities[i] = Mathf.Pow(1f / (occurencies[i] + 1), 2);

            float sumProb = probabilities.Sum();

            for (int i = 0; i < probabilities.Length; i++)
                probabilities[i] /= sumProb;

            float[] cumulativeProbabilities = new float[The_Helper.EnumCount<Potions>()];

            cumulativeProbabilities[0] = probabilities[0];
            for (int i = 1; i < cumulativeProbabilities.Length; i++)
                cumulativeProbabilities[i] = cumulativeProbabilities[i - 1] + probabilities[i];

            float r = Random.Range(0, 1f);

            for (int i = 0; i < cumulativeProbabilities.Length; i++)
                if (r <= cumulativeProbabilities[i])
                {
                    potion = (Potions)i;
                    break;
                }

            Debug.Log("Choice: " + potion + " | r: " + r + "\n" + cumulativeProbabilities.ToReadableString());
        }

        lastPotions.Enqueue(potion);

        return potion;
	}

	private GameObject CreatePotionSprite(Potions newPot)
	{
		GameObject newSprite = null;
		switch(newPot){
		case Potions.Fire:
			newSprite = GameObject.Instantiate( FireSprite );
			break;

		case Potions.Nature:
			newSprite = GameObject.Instantiate( NatureSprite );
			break;

		case Potions.Water:
			newSprite = GameObject.Instantiate( WaterSprite );
			break;
		}
		return newSprite;
	}

	private IEnumerator StartConveyorAnimation( Potions newPot )
	{
		GameObject newSprite = CreatePotionSprite(newPot);


		if( moveToAnimator.transform.childCount > 0 )
		{
			var childT = moveToAnimator.transform.GetChild(0);
			childT.SetParent( moveToAnimator.transform.parent);
			StartCoroutine( DrunkPotionAnimationCooroutine(childT) );
		}

		if( getInAnimator.transform.childCount > 0 )
		{
			var child = getInAnimator.transform.GetChild(0);
			child.transform.SetParent( moveToAnimator.transform );
			child.transform.localScale = Vector3.one;
			child.transform.localPosition = Vector3.one;

			moveToAnimator.SetTrigger("Move");
		}

		newSprite.transform.SetParent(getInAnimator.gameObject.transform);
		newSprite.transform.localScale = Vector3.one;
		newSprite.transform.localPosition = Vector3.one;
		getInAnimator.SetTrigger("Move");

		yield return new WaitForSeconds(1f);

	}

	private IEnumerator DrunkPotionAnimationCooroutine(Transform child )
	{
		while(child.localScale.x> 0.1)
		{
			child.localScale = Vector3.one * (child.localScale.x-Time.deltaTime  );
			child.transform.position = child.transform.position + Vector3.up * Time.deltaTime;
			yield return null;
		}

		Destroy( child.gameObject );
	}

	public void EmptySuckFill()
	{
		StartCoroutine(Empty());
	}

	private IEnumerator Empty()
	{
        yield return new WaitForSeconds(.65f);
		while( SuckFillImage.fillAmount >0 )
		{
			SuckFillImage.fillAmount -= Time.deltaTime * 2.5f;
			yield return null;
		}
		SuckFillImage.fillAmount=0;
	}

	private IEnumerator SuckPotionAnimationCooroutine(Potions pot )
	{
		SuckFillImage.fillAmount = 0;
		SuckFillImage.color = LaserLogic.GetColorOfPotion(pot).Change(ColorProperty.a, .65f);
		while( SuckFillImage.fillAmount <1.0 )
		{
			SuckFillImage.fillAmount += Time.deltaTime * 1.5f;
			yield return null;
		}
		//SuckCurrentImage.color = SuckFillImage.color;
		//SuckFillImage.fillAmount=0;
	}

}
