﻿using UnityEngine;
using System.Collections;

/// <summary>
/// Base class for minor animations - Provides an Interpolate function which
/// can be overriden. Its input moves from 0 to 1 back to 0.
/// </summary>
public class PropAnimation : MonoBehaviour
{  
    /// <summary>
    /// Time it takes to interpolate from start to finish and back
    /// </summary>
    public Vector2 oneWayTripTimeMinMax = Vector2.one;

    public float pauseStart = 0;
    public float pauseEnd = 0;
    public bool autoInit = true; 
    public bool randomStart = false;

    float oneWayTripTime = 1;
    float fullLength { get { return 2 * oneWayTripTime + pauseStart + pauseEnd; } }
    bool paused = false; 
    float timePassed = 0;

    public void Awake()
    {

        if (randomStart)
            timePassed = randomStart ? Random.Range(0, fullLength) : 0;

        oneWayTripTime = oneWayTripTimeMinMax.RandomBetween(); 
        if (autoInit)
            Initialize();
    }
    // Update is called once per frame
    void Update()
    {
        if (paused)
            return;

        if (oneWayTripTime == 0)
            oneWayTripTime = 1;


        float timePassed_Circular = timePassed % fullLength;
        timePassed += Time.deltaTime;
        float interpolator = 0;

        // First Pause
        if (timePassed_Circular < pauseStart)
            return;
        // 0..1
        else if (timePassed_Circular < pauseStart + oneWayTripTime) 
            interpolator = (timePassed_Circular - pauseStart) / (oneWayTripTime);  
        // Second Pause
        else if (timePassed_Circular < pauseStart + oneWayTripTime + pauseEnd)
            return;
        // 1..0
        else 
            interpolator = 1 - (timePassed_Circular - pauseStart - oneWayTripTime - pauseEnd) / oneWayTripTime; 
                  
        Interpolate(interpolator);  

    }
  

    public void Pause()
    {
        paused = true;
    }

    public void Play()
    {
        paused = false;
    }

    public void Reset()
    {
        Awake();
    }

    void OnDisable()
    {
        Pause();
    }

    void OnEnable()
    {
        Play();

    }
    protected virtual void Interpolate(float t)
    {

    }
    protected virtual void Initialize()
    {

    }
}

