﻿using UnityEngine;
using System.Collections;

public class TransformAnimation : PropAnimation
{

    public DisplaceType move = DisplaceType.BackAndForth;
    public Vector3 displacement = Vector3.up;

    public DisplaceType spin = DisplaceType.No;
    public Vector3 rotation = Vector3.zero;
    
    /// <summary>
    /// Is the current setup the Start, Middle or Finish of the animation?
    /// </summary>
    public PositionType positionType = PositionType.Middle;


    Vector3 initPos;
    Quaternion initRot; 
    Rigidbody myRigidbody;
    float lastT;
    
    protected override void Initialize()
    {

        myRigidbody = GetComponent<Rigidbody>();

        if (positionType == PositionType.Middle)
        {
            transform.localPosition -= displacement / 2;
            transform.localRotation *= Quaternion.Inverse(Quaternion.Euler(rotation / 2));
        }
        else if (positionType == PositionType.End)
        {
            transform.localPosition -= displacement;
            transform.localRotation *= Quaternion.Inverse(Quaternion.Euler(rotation));
        }

        initPos = transform.localPosition;
        initRot = transform.localRotation;
    }
    protected override void Interpolate(float t)
    { 
        if (move == DisplaceType.BackAndForth)
            LocalMove(Vector3.Lerp(initPos, initPos + transform.localRotation * displacement, t));
        else if (move == DisplaceType.OneWay)
            LocalMove(transform.localPosition + transform.localRotation * displacement * Mathf.Abs(t - lastT));

        if (spin == DisplaceType.BackAndForth)
            LocalSpin(Quaternion.Lerp(initRot, initRot * Quaternion.Euler(rotation), t));
        else if (spin == DisplaceType.OneWay)
            LocalSpin(Quaternion.Lerp(transform.localRotation, transform.localRotation * Quaternion.Euler(rotation), Mathf.Abs(t-lastT)));
        //LocalSpin(transform.localRotation * Quaternion.Euler(rotation initPos) * Mathf.Abs(t - lastT));

        lastT = t;
    }

    void LocalMove(Vector3 pos)
    {
        Vector3 localPos = pos;
        if (transform.parent)
            localPos = transform.parent.TransformPoint(pos);

        if (myRigidbody)
            myRigidbody.MovePosition(localPos);
        else
            transform.position = localPos;
    }

    void LocalSpin(Quaternion localRot)
    {
        Quaternion absoluteRot = localRot;
        if (transform.parent)
            absoluteRot = transform.parent.rotation * localRot;
        if (myRigidbody)
            myRigidbody.MoveRotation(absoluteRot);
        else
            transform.rotation = absoluteRot;
    }
    public enum PositionType { Start = 0, Middle, End }
    public enum DisplaceType { No = 0, OneWay, BackAndForth }
}
